// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "fw/base.h"
#include "fw/lewan/config.h"
#include "fw/lewan/context.h"
#include "fw/lewan/conv.h"
#include "fw/servo/visitor.h"
#include "protocol.h"

#include <emlabcpp/pid.h>
#include <schpin_lib/algorithm.h>

#pragma once

namespace schpin
{

struct lewan_goto_event
{
        lewan_angle g;
        btime       t;
};

struct lewan_spin_event
{
        lewan_angular_velocity av;
        btime                  t;
};

struct lewan_config_event
{
        lewan_config_enum addr;
        config_buffer     buffr;
};

using lewan_servo_events_variant =
    std::variant< lewan_goto_event, lewan_spin_event, lewan_config_event >;

class lewan_servo
{
        struct activate_task
        {
                std::size_t i = 0;
                cmd_roulette<
                    LEWAN_SERVO_ANGLE_OFFSET_READ,
                    LEWAN_SERVO_ANGLE_LIMIT_READ,
                    LEWAN_SERVO_VIN_LIMIT_READ,
                    LEWAN_SERVO_TEMP_MAX_LIMIT_READ,
                    LEWAN_SERVO_OR_MOTOR_MODE_READ,
                    LEWAN_SERVO_LED_CTRL_READ,
                    LEWAN_SERVO_POS_READ >
                    startup_roulette;
        };

        struct goto_task
        {
                lewan_full_angle start;
                lewan_angle      goal;
                clk_time         start_time;
                btime            time;
        };

        struct spin_task
        {
                clk_time clock_end;
        };

        struct idle_task
        {
        };

        using task_variant = std::variant< activate_task, goto_task, spin_task, idle_task >;

        lewan_servo_id                        id_;
        lewan_config_map                      config_map_;
        lewan_context                         context_;
        lewan_converter                       converter_;
        task_variant                          task_ = { activate_task{} };
        lewan_full_angle                      last_report_;
        message_catcher< lewan_master_group > lewan_catcher_;
        std::size_t                           poll_i_         = 0;
        s_ctl_id                              msg_id_counter_ = 0;

        bool                new_pos_          = false;
        bool                vel_pid_enabled_  = false;
        float               desired_velocity_ = 0.f;
        em::pid< clk_time > vel_pid_;

        bool                pos_pid_enabled_ = false;
        float               desired_pos_     = 0.f;
        em::pid< clk_time > pos_pid_;

public:
        using receive_type   = lewan_master_variant;
        using events_variant = lewan_servo_events_variant;

        lewan_servo( lewan_servo_id id, clk_time now )
          : id_( id )
          , converter_( config_map_ )
          , vel_pid_( now )
          , pos_pid_( now )
        {
                configure_pid();
        }

        servo_id get_id() const
        {
                return *id_;
        }

        bool has_task() const
        {
                return !std::holds_alternative< idle_task >( task_ );
        }

        bool is_activating() const
        {
                return std::holds_alternative< activate_task >( task_ );
        }

        const lewan_servo_state& get_state() const
        {
                return *context_;
        }

        void stop( servo_visitor& );

        std::optional< servo_master_message >
        on_servo_receive( clk_time, s_ctl_id, const lewan_master_variant& );

        void tick( clk_time, servo_visitor& );

        void on_event( ctl_id, clk_time, const lewan_servo_events_variant&, servo_visitor& );

        void poll( clk_time, std::size_t, servo_visitor& );

        em::either< lewan_servo_events_variant, servo_master_message >
        on_servo_packet( ctl_id id, master_servo_command_variant subcmd );

        void dump_info();

private:
        void activate_tick( activate_task s, servo_visitor& v )
        {
                switch ( s.i ) {
                        case 0:
                                s.startup_roulette.tick( [&]( auto tag ) {  //
                                        transmit_servo< decltype( tag )::value, reply::YES >( v );
                                } );
                                if ( s.startup_roulette.get_index() != 0 ) {
                                        task_ = s;
                                        return;
                                }
                                break;
                        case 1:
                                lewan_catcher_.set_trap( LEWAN_SERVO_POS_READ );
                                transmit_servo< LEWAN_SERVO_POS_READ, reply::YES >( v );
                                break;
                        case 2:
                                if ( !lewan_catcher_.catched() ) {
                                        return;
                                }
                                break;
                        default:
                                v.on_activate( *id_ );
                                task_ = idle_task{};
                                return;
                }

                s.i += 1;
                task_ = s;
        }

        template < messages_enum ID, typename... Args >
        void servo_reply_ctl( servo_visitor& v, Args&&... args )
        {
                v.to_control( *id_, servo_master_group::create< ID >( args... ) );
        }

        template < error_protocol_enum ID, typename... Args >
        void servo_reply_ctl_error( ctl_id cid, servo_visitor& v, Args&&... args )
        {
                servo_reply_ctl< ERROR >( v, cid, servo_error_group::create< ID >( args... ) );
        }

        template < lewan_message_enum ID, reply Rpl, typename... Args >
        void transmit_servo( servo_visitor& v, Args... args )
        {
                msg_id_counter_++;
                v.to_servo( msg_id_counter_, Rpl, make_master_lewan_message< ID >( id_, args... ) );
        }

        void configure_pid()
        {
                vel_pid_.set_config( em::pid< clk_time >::config{
                    config_map_.get_val< LEWAN_VEL_P_GAIN >().as_float(),
                    config_map_.get_val< LEWAN_VEL_I_GAIN >().as_float(),
                    config_map_.get_val< LEWAN_VEL_D_GAIN >().as_float(),
                    -1000.f,
                    1000.f } );

                pos_pid_.set_config( em::pid< clk_time >::config{
                    config_map_.get_val< LEWAN_GOTO_P_GAIN >().as_float(),
                    config_map_.get_val< LEWAN_GOTO_I_GAIN >().as_float(),
                    config_map_.get_val< LEWAN_GOTO_D_GAIN >().as_float(),
                    -1000.f,
                    1000.f } );
        }
};

}  // namespace schpin
