#include "fw/dynmx/serial.h"
#include "fw/serial.h"
#include "fw/servo/holder.h"
#include "node/lua_vm/extra_ser.h"
#include "test/servo_runner/lua_def.h"

#include <schpin_lib/bucket_memory_resource.h>
#include <schpin_lib/json/tuple_variant_serializer.h>
#include <schpin_lib/lua/help.h>
#include <schpin_lib/lua/repl.h>
#include <schpin_lib/lua/util.h>
#include <schpin_lib/lua/vm.h>
#include <schpin_lib_exp/config.h>

#pragma once

template <>
struct nlohmann::adl_serializer< schpin::master_servo_command_variant >
  : schpin::tuple_variant_serializer< schpin::master_servo_command_variant >
{
};

template <>
struct nlohmann::adl_serializer< schpin::servo_master_command_variant >
  : schpin::tuple_variant_serializer< schpin::servo_master_command_variant >
{
};

template <>
struct nlohmann::adl_serializer< schpin::servo_error_variant >
  : schpin::tuple_variant_serializer< schpin::servo_error_variant >
{
};

namespace schpin
{

template <>
struct ros_param_serializer< servo_type > : ros_param_serializer_enum< servo_type >
{
};

using runner_error = error< struct runner_error_tag >;

struct servo_tester_config
{
        servo_type                             type;
        std::optional< servo_id >              opt_id;
        std::optional< std::filesystem::path > opt_dev;
        std::optional< std::filesystem::path > opt_source;

        static auto ros_param_get()
        {
                return std::make_tuple(  //
                    ros_param< servo_type >( "stype", "Servo type" ),
                    ros_param< std::optional< servo_id > >( "sid", "Servo id" ),
                    ros_param< std::optional< std::filesystem::path > >(
                        "device", "Device to open for connection" ),
                    ros_param< std::optional< std::filesystem::path > >(
                        "script", "Script to execute instead of the interactive mode" ) );
        }
};

class servo_runner
{
        using mem              = bucket_memory_resource< 1, servo_bucket_size >;
        using deleter          = typename mem::deleter< servo_holder >;
        using servo_holder_ptr = std::unique_ptr< servo_holder, deleter >;

        mem                                memory_;
        std::optional< dynmx_serial >      dynmx_ser_;
        std::optional< serial_connection > lewan_ser_;
        servo_holder                       servo_;
        slua_vm< slua_runner_def >         vm_;
        slua_repl                          repl_;

        struct visitor : servo_visitor
        {
                servo_runner& runner;

                visitor( servo_runner& srunner )
                  : runner( srunner )
                {
                }

                void to_control( servo_id, const servo_master_message& )
                {
                }

                void on_activate( servo_id id )
                {
                        SCHPIN_INFO_LOG( "lewan_servo", "Got activation of: " << id );
                }

                void to_servo( s_ctl_id msg_id, reply r, const master_lewan_message& m )
                {
                        SCHPIN_ASSERT( runner.lewan_ser_ );

                        runner.lewan_ser_->write( m );
                        if ( r == reply::NO ) {
                                return;
                        }
                        runner.receive_serial_msg( msg_id );
                }

                void to_servo( s_ctl_id msg_id, reply, const master_dynmx_message& m )
                {
                        SCHPIN_ASSERT( runner.dynmx_ser_ );

                        runner.dynmx_ser_->send( m );

                        dynmx_master_message msg = runner.dynmx_ser_->read_reply();

                        dynmx_master_protocol::extract( msg ).match(
                            em::uncurry |
                                [&]( dynmx_header, dynmx_servo_id, auto tpl, dynmx_crc ) {
                                        const dynmx_response_error& err = std::get< 1 >( tpl );
                                        if ( err != dynmx_no_error ) {
                                                SCHPIN_ERROR_LOG(
                                                    "runner", "Dynmx return error: " << err );
                                        }
                                        const dynmx_buffer& buffer = std::get< 2 >( tpl );

                                        if ( buffer.size() != 0 ) {
                                                runner.servo_.on_servo_receive(
                                                    get_system_time(), msg_id, buffer );
                                        }
                                },
                            [&]( auto err ) {
                                    SCHPIN_ERROR_LOG( "servo_runner", err << ", msg: " << msg );
                            } );
                }
        };

public:
        servo_runner(
            const servo_tester_config&         conf,
            std::optional< serial_connection > opt_ser = {} )
          : servo_( *servo_holder::make( conf.type, *conf.opt_id, get_system_time(), &memory_ ) )
        {
                SCHPIN_ASSERT( conf.opt_id );
                if ( conf.type == DYNMX ) {
                        SCHPIN_ASSERT( opt_ser );
                        dynmx_ser_.emplace( *opt_ser );
                }
                visitor vis{ *this };
                while ( servo_.is_activating() ) {
                        servo_.tick( get_system_time(), vis );
                }
        }

        void receive_serial_msg( s_ctl_id msg_id )
        {
                static constexpr std::size_t n = lewan_master_message::max_size;

                SCHPIN_ASSERT( lewan_ser_ );

                lewan_ser_->read< n >().match(
                    [&]( const em::static_vector< uint8_t, n >& data ) {
                            lewan_sequencer< lewan_master_message > sequencer;

                            bool load_fired = sequencer.load( data, [&]( lewan_master_message m ) {
                                    parse_message( msg_id, m );
                            } );

                            if ( !load_fired ) {
                                    SCHPIN_ERROR_LOG(
                                        "control",
                                        "servo not replying, got:" << convert_view< int >( data ) );
                            }
                    },
                    [&]( auto err ) {
                            SCHPIN_ERROR_LOG( "control", "Read error: " << err );
                    } );
        }
        void parse_message( s_ctl_id msg_id, const lewan_master_message& m )
        {
                lewan_master_protocol::extract( m ).match(
                    em::uncurry |
                        [&]( lewan_header_start,
                             lewan_header_start,
                             lewan_servo_id,
                             lewan_master_variant msg_var,
                             lewan_checksum ) {
                                auto clk_time = get_system_time();
                                SCHPIN_INFO_LOG( "servo", "clktime: " << clk_time );
                                servo_.on_servo_receive( clk_time, msg_id, msg_var );
                        },
                    [&]( auto err ) {
                            SCHPIN_ERROR_LOG( "lewan_servo", "Lewan prot error: " << err );
                    } );
        }

        void tick()
        {
                visitor vis{ *this };
                servo_.tick( get_system_time(), vis );
                servo_.poll( get_system_time(), 0, vis );
                std::string msg = repl_.prompt();
                vm_.load( msg ).optionally_log();
                vm_.full_exec( [&]( lua_State* lua_state, auto event ) {
                           return handle_event( lua_state, event );
                   } )
                    .optionally_log();
        }

        void run_file( const std::filesystem::path& file )
        {
                auto opt_e = vm_.load( file );
                if ( opt_e ) {
                        opt_e.optionally_log();
                        return;
                }
                bool finished = false;

                visitor vis{ *this };
                while ( !finished ) {
                        servo_.tick( get_system_time(), vis );
                        servo_.poll( get_system_time(), 0, vis );
                        vm_.exec( [&]( lua_State* lua_ptr, auto event ) {
                                return handle_event( lua_ptr, event );
                        } )
                            << match(
                                   [&]( lua_res res ) {
                                           finished = res == lua_res::FINISHED;
                                   },
                                   [&]( slua_error err ) {
                                           finished = true;
                                           SCHPIN_ERROR_LOG( "runner", "Slua error: " << err );
                                   } );
                }
        }

        void handle_event( lua_State*, print_help_event )
        {
                nlohmann::json res = slua_runner_def::get_def();
                print_help( std::cout, res );
        }

        void handle_event( lua_State*, quit_event e )
        {
                std::exit( e.res );
        }

        void handle_event( lua_State*, print_config_event )
        {
                servo_.dump_info();
        }

        nlohmann::json handle_event( lua_State* lua_ptr, msg_event e )
        {
                nlohmann::json res;

                // TODO: write proper thing for this in schpin_lib/json and check in rest of the
                // codebase
                try {
                        auto var = e.j.get< master_servo_command_variant >();

                        servo_.on_servo_packet( ctl_id{ 0 }, var )
                            .match(
                                [&]( servo_event se ) {
                                        visitor vis{ *this };
                                        servo_.on_event(
                                            ctl_id{ 0 }, get_system_time(), se.sub_event, vis );
                                },
                                [&]( servo_master_message m ) {
                                        servo_master_group::extract( m ).match(
                                            [&]( auto var ) {
                                                    res = var;
                                            },
                                            [&]( auto err ) {
                                                    SCHPIN_ERROR_LOG(
                                                        "robot",
                                                        "this should not happen error: " << err );
                                            } );
                                } );
                }
                catch ( const nlohmann::json::exception& ex ) {
                        lua_error_stream( lua_ptr, "Json error: ", ex.what() );
                }
                return res;
        }
};

}  // namespace schpin
